const { EventEmitter } = require('events');
const { Logger } = require('twlv-logger');
const assert = require('assert');
const { hidden } = require('./prop');
const WebSocket = process.browser ? window.WebSocket : require('ws');
const { Request } = require('./request');
const { Identity } = require('twlv-identity');

class Client extends EventEmitter {
  /**
   * Constructor
   *
   * @param {string} url twlv daemon ipc url
   */
  constructor (url) {
    super();

    assert(typeof url === 'string', 'Client require string url as parameter');

    this.url = url;
    this.identity = undefined;
    this.networkId = '';
    this.advertisement = undefined;

    hidden(this, 'requests', []);
    hidden(this, 'reqHandlers', {});
    hidden(this, 'messageHandlers', {});
    hidden(this, 'logger', new Logger('twlv-client'));
    hidden(this, 'sock');
    hidden(this, 'module');
  }

  get address () {
    return this.identity.address;
  }

  /**
   * Connected status
   */
  get connected () {
    return Boolean(this.sock && this.sock.readyState === WebSocket.OPEN);
  }

  /**
   * Connecting status
   */
  get connecting () {
    return Boolean(this.sock && this.sock.readyState === WebSocket.CONNECTING);
  }

  /**
   * Connect to twlv daemon, when module specified client will connect to daemon
   * as module and daemon will add new external module
   *
   * @param {object?} module Module definition { kind, name, version } (optional)
   */
  async connect (module) {
    if (this.connecting || this.connected) {
      return;
    }

    await new Promise((resolve, reject) => {
      let sock = this.sock = new WebSocket(`${this.url}/ipc`);
      sock.onopen = () => {
        this.logger.log('IPC Client connecting....');
        this.write('connect', module);
      };

      sock.onmessage = evt => {
        let { command, payload } = JSON.parse(evt.data);
        if (command !== '!connect') {
          sock.close();
          return;
        }

        let { address, privateKey, networkId, advertisement } = payload;
        this.identity = new Identity(privateKey);
        if (address !== this.identity.address) {
          sock.close();
          return;
        }
        this.networkId = networkId;
        this.advertisement = advertisement;

        sock.onmessage = this._onWireMessage.bind(this);

        resolve();
      };

      sock.onclose = () => {
        this.logger.log('Sock closed');
        this.sock = undefined;
      };
    });
  }

  /**
   * Disconnect client from daemon
   */
  disconnect () {
    this.logger.log('Client disconnecting ...');
    if (this.sock) {
      this.sock.close();
    }
  }

  /**
   * Write to daemon
   *
   * @param {string} command
   * @param {var} payload
   */
  write (command, payload) {
    this.sock.send(JSON.stringify({ command, payload }));
  }

  broadcast ({ command, payload }) {
    assert(typeof command === 'string', 'Broadcast must specify command as string');

    this.write('broadcast', { command, payload });
  }

  send ({ address, command, payload }) {
    assert(typeof address === 'string', 'Send must specify address as string');
    assert(typeof command === 'string', 'Send must specify command as string');

    let req = new Request(this._removeRequest.bind(this));
    this._addRequest(req);
    return req.run(req => {
      this.write('send', { id: req.id, address, command, payload });
    });
  }

  relay ({ address, command, payload }) {
    assert(typeof address === 'string', 'Relay must specify address as string');
    assert(typeof command === 'string', 'Relay must specify command as string');

    this.write('relay', { address, command, payload });
  }

  lookup (address) {
    assert(typeof address === 'string', 'Lookup must specify address as string');

    let req = new Request(this._removeRequest.bind(this));
    this._addRequest(req);
    return req.run(() => {
      this.write('lookup', { id: req.id, address });
    });
  }

  request ({ address, command, payload, timeout }) {
    assert(typeof address === 'string', 'Request must specify address as string');
    assert(typeof command === 'string', 'Request must specify command as string');

    let req = new Request(this._removeRequest.bind(this));
    this._addRequest(req);
    return req.run(req => {
      this.write('request', { id: req.id, address, command, payload });
    });
  }

  /**
   * This method is supposed to invoke internally
   * Several event commands supported:
   * - !send
   * - send
   * - !lookup
   * - lookup
   * - message
   * - advertisement
   *
   * @param {Event} evt Websocket event
   */
  async _onWireMessage (evt) {
    let { command, payload: wirePayload } = JSON.parse(evt.data);
    this.logger.log('IPC<client>: %o', command);

    if (command === '!send') {
      let { id, error } = wirePayload;
      let req = this.requests.find(req => req.id === id);
      if (req) {
        if (error) {
          req.reject(new Error(error));
        } else {
          req.resolve();
        }
      }
      return;
    }

    if (command === '!lookup') {
      let { id, value, error } = wirePayload;
      let req = this.requests.find(req => req.id === id);
      if (req) {
        if (error) {
          req.reject(new Error(error));
        } else {
          req.resolve(value);
        }
      }
      return;
    }

    if (command === 'message') {
      let { address, command, payload } = wirePayload;

      if (command === '!req') {
        let { id, command, payload: param } = payload;
        let handler = this.reqHandlers[command];
        if (!handler) {
          this.logger.warn(`Request handler for ${command} not available`);
          return;
        }
        try {
          let value = await handler(param, address);
          this.send({ address, command: '!res', payload: { id, value } });
        } catch (err) {
          this.logger.error('Caught on onRequest', err);
          let error = err.message;
          this.send({ address, command: '!res', payload: { id, error } });
        }
      } else if (command === '!res') {
        let { id, value, error } = wirePayload.payload;
        let req = this.requests.find(req => req.id === id);
        if (req) {
          if (error) {
            req.reject(new Error(`IPC Request: ${error}`));
          } else {
            req.resolve(value);
          }
        }
      } else {
        let handler = this.messageHandlers[command];
        if (!handler) {
          this.logger.warn(`Message handler for ${command} not available`);
          return;
        }
        try {
          await handler(wirePayload);
        } catch (err) {
          this.logger.error('Caught on onMessage', err);
        }
      }
      return;
    }

    this.emit(command, wirePayload);
  }

  _removeRequest (req) {
    let index = this.requests.indexOf(req);
    if (index !== -1) {
      this.requests.splice(index, 1);
    }
  }

  _addRequest (req) {
    this.requests.push(req);
  }

  setRequestHandler (command, callback) {
    this.reqHandlers[command] = callback;
  }

  setMessageHandler (command, callback) {
    this.messageHandlers[command] = callback;
  }
}

module.exports = { Client };
